/* Copyright (c) 2005, David A. Clunie DBA Pixelmed Publishing. All rights reserved. */

header {
	package com.pixelmed.ucum;
}

class UCUMRewriter extends TreeParser;

options {
        buildAST=true;
}

{
	int mode = 0;				// 0 = not set, 1 = rewrite binary trees, 2 = flatten trees
	
	void setMode(int mode)			{ this.mode = mode; }
	
	int debugMode = 0;

	void setDebugMode(int debugMode)	{ this.debugMode = debugMode; }
	
	class OurNodeCloner {
		AST clone(AST x) {
			AST xCopy = astFactory.dup(x);
			AST child = x.getFirstChild();
			while (child != null) {
				if (debugMode > 2) { System.err.println("value: duplicating child of x = "+child.toStringList()); }
				xCopy.addChild(clone(child));
				child = child.getNextSibling();
			}
			return xCopy;
		}
	};
	OurNodeCloner ourNodeCloner = new OurNodeCloner();
	
	double ourvalueofpi = java.lang.Math.PI;	// to prevent token name replacement in action
}

expr:!		#(m:MULTIPLY lm:expr rm:expr)		{ if (mode == 1) {
								// binary tree rewriting
								if (debugMode > 1) { System.err.println("expr: multiply l = "+ #lm.toStringList()); }
								if (debugMode > 1) { System.err.println("expr: multiply r = "+ #rm.toStringList()); }
								if (#lm.getType() == DIVIDE) {
									// make (ll/lr).r into (ll.r)/lr
									AST newr = ourNodeCloner.clone(#rm);
									AST newll = ourNodeCloner.clone(#lm.getFirstChild());
									AST newlr = ourNodeCloner.clone(#lm.getFirstChild().getNextSibling());
									#expr = #(#[DIVIDE,"/"], #(#[MULTIPLY,"."], newll, newr), newlr);
								}
								else if (#rm.getType() == DIVIDE) {
									// make l.(rl/rr) into (l.rl)/rr
									AST newl = ourNodeCloner.clone(#lm);
									AST newrl = ourNodeCloner.clone(#rm.getFirstChild());
									AST newrr = ourNodeCloner.clone(#rm.getFirstChild().getNextSibling());
									#expr = #(#[DIVIDE,"/"], #(#[MULTIPLY,"."], newl, newrl), newrr);
								}
								else if (#lm.getType() == MULTIPLY && #lm.getFirstChild().getType() == DIGITS && #lm.getFirstChild().getNextSibling().getType() != DIGITS) {
									// make (dll.lr).r into dll.(lr.r) (propagate constants upwards)
									if (debugMode > 1) { System.err.println("propagate constants upwards"); }
									AST newr = ourNodeCloner.clone(#rm);
									AST newll = ourNodeCloner.clone(#lm.getFirstChild());
									AST newlr = ourNodeCloner.clone(#lm.getFirstChild().getNextSibling());
									#expr = #(#[MULTIPLY,"."], newll, #(#[MULTIPLY,"."], newlr, newr));
								}
								else if (#lm.getType() == DIGITS && #rm.getType() == MULTIPLY && #rm.getFirstChild().getType() == DIGITS) {
									if (debugMode > 1) { System.err.println("collapse first constant multiplication"); }
									double dl = Double.parseDouble(#lm.getText());
									double drl = Double.parseDouble(#rm.getFirstChild().getText());
									AST newrr = ourNodeCloner.clone(#rm.getFirstChild().getNextSibling());
									#expr = #(#[MULTIPLY,"."], #[DIGITS,Double.toString(dl*drl)], newrr);
								}
								
								else if (#lm.getType() == DIGITS && #rm.getType() == DIGITS) {
									if (debugMode > 1) { System.err.println("collapse constant multiplication: #lm.getText() = "+ #lm.getText()); }
									if (debugMode > 1) { System.err.println("collapse constant multiplication: #rm.getText() = "+ #rm.getText()); }
									double dl = Double.parseDouble(#lm.getText());
									double dr = Double.parseDouble(#rm.getText());
									#expr = #[DIGITS,Double.toString(dl*dr)];
								}
								else if (#lm.getType() == DIGITS && #rm.getType() != DIGITS) {
									// sort digits to the left (multiplication is commutative)
									if (debugMode > 1) { System.err.println("expr: multiply: leave left digits and right non-digits"); }
									#expr = #(#m, #lm, #rm);
								}
								else if (#lm.getType() != DIGITS && #rm.getType() == DIGITS) {
									// sort digits to the left (multiplication is commutative)
									if (debugMode > 1) { System.err.println("expr: multiply: exchange left non-digits with right digits"); }
									#expr = #(#m, #rm, #lm);
								}
								else if (#lm.getFirstChild() == null && #rm. getFirstChild() != null) {
									// sort trees as opposed to leaves to the left (multiplication is commutative)
									if (debugMode > 1) { System.err.println("expr: multiply: exchange left leaf with right tree"); }
									#expr = #(#m, #rm, #lm);
								}
								else if (#lm.getFirstChild() != null && #rm. getFirstChild() == null) {
									// sort trees as opposed to leaves to the left (multiplication is commutative)
									if (debugMode > 1) { System.err.println("expr: multiply: leave left tree with right leaf"); }
									#expr = #(#m, #lm, #rm);
								}
								else if (#lm.getType() < #rm.getType()) {
									// sort by type (multiplication is commutative)
									if (debugMode > 1) { System.err.println("expr: multiply: exchange left with right based on type"); }
									#expr = #(#m, #rm, #lm);
								}
								else {
									#expr = #(#m, #lm, #rm);
								}
								if (debugMode > 1) { System.err.println("expr: multiply result = "+ #expr.toStringList()); }
							  }
							  else if (mode == 2) {
								// flatten binary tree into n-tree
								if (debugMode > 1) { System.err.println("expr: multiply l = "+ #lm.toStringList()); }
								if (debugMode > 1) { System.err.println("expr: multiply r = "+ #rm.toStringList()); }
								
								if (#rm.getNextSibling() == null) {
									// make (ll.lr).(rl.rr) into ll.lr.rl.rr
									if (debugMode > 1) { System.err.println("flatten left and right multiply"); }
									AST new_node = #(#[MULTIPLY,"."]);
									AST child = #lm.getFirstChild();
									if (child == null) {
										new_node.addChild(ourNodeCloner.clone(#lm));
									}
									else {
										while (child != null) {
											AST newChild = ourNodeCloner.clone(child);
											new_node.addChild(newChild);
											child=child.getNextSibling();
										}
									}
									child = #rm.getFirstChild();
									if (child == null) {
										new_node.addChild(ourNodeCloner.clone(#rm));
									}
									else {
										while (child != null) {
											AST newChild = ourNodeCloner.clone(child);
											new_node.addChild(newChild);
											child=child.getNextSibling();
										}
									}
									#expr = new_node;
								}
								if (debugMode > 1) { System.err.println("expr: multiply result = "+ #expr.toStringList()); }
							  }
							}
	|!	#(d:DIVIDE l:expr r:expr)		{ if (mode == 1) {
								// binary tree rewriting
								if (debugMode > 1) { System.err.println("expr: divide l = "+ #l.toStringList()); }
								if (debugMode > 1) { System.err.println("expr: divide r = "+ #r.toStringList()); }
								if (#l.getType() == DIVIDE) {
									// make (ll/lr)/r into ll/(lr.r)
									if (debugMode > 1) { System.err.println("rewrite divide case 1: l = "+ #l.toStringList()); }
									if (debugMode > 1) { System.err.println("rewrite divide case 1: r = "+ #r.toStringList()); }
									AST newr = ourNodeCloner.clone(#r);
									if (debugMode > 1) { System.err.println("rewrite divide case 1: newr = "+ newr.toStringList()); }
									AST newll = ourNodeCloner.clone(#l.getFirstChild());
									if (debugMode > 1) { System.err.println("rewrite divide case 1: newll = "+ newll.toStringList()); }
									AST newlr = ourNodeCloner.clone(#l.getFirstChild().getNextSibling());
									if (debugMode > 1) { System.err.println("rewrite divide case 1: newlr = "+ newlr.toStringList()); }
									#expr = #(#[DIVIDE,"/"], newll, #(#[MULTIPLY,"."], newlr, newr));
								}
								else if (#r.getType() == DIVIDE) {
									// make l/(rl/rr) into l.rr/rl
									AST newl = ourNodeCloner.clone(#l);
									AST newrl = ourNodeCloner.clone(#r.getFirstChild());
									AST newrr = ourNodeCloner.clone(#r.getFirstChild().getNextSibling());
									#expr = #(#[DIVIDE,"/"], #(#[MULTIPLY,"."], newl, newrr), newrl);
								}
								else if (#l.getType() == MULTIPLY && #r.getType() == MULTIPLY
								     && #l.getFirstChild().getType() == DIGITS && #r.getFirstChild().getType() == DIGITS) {
									double dll = Double.parseDouble(#l.getFirstChild().getText());
									double drl = Double.parseDouble(#r.getFirstChild().getText());
									AST newlr = ourNodeCloner.clone(#l.getFirstChild().getNextSibling());
									AST newrr = ourNodeCloner.clone(#r.getFirstChild().getNextSibling());
									#expr = #(#[MULTIPLY,"."], #[DIGITS,Double.toString(dll/drl)], #(#[DIVIDE,"/"], newlr, newrr));
								}
								else if (#l.getType() == MULTIPLY && #r.getType() == DIGITS
								     && #l.getFirstChild().getType() == DIGITS) {
									double dll = Double.parseDouble(#l.getFirstChild().getText());
									double dr = Double.parseDouble(#r.getText());
									AST newlr = ourNodeCloner.clone(#l.getFirstChild().getNextSibling());
									#expr = #(#[MULTIPLY,"."], #[DIGITS,Double.toString(dll/dr)], newlr);
								}
								else {
									#expr = #(#d, #l, #r);
								}
								if (debugMode > 1) { System.err.println("expr: divide result = "+ #expr.toStringList()); }
							  }
							  else if (mode == 2) {
								#expr = #(#d, #l, #r);
							  }
							}
	|!	#(EVAL_ATOM e:expr)	{ #expr = e; }
	|!	TEN			{ #expr = #([DIGITS,"10"]) ; }
	|!	TENALT			{ #expr = #([DIGITS,"10"]) ; }
	|!	PI			{ #expr = #([DIGITS,Double.toString(ourvalueofpi)]) ; }
	|!	PERCENT			{ #expr = #([DIGITS,"1E-2"]) ; }
	|!	PPTH			{ #expr = #([DIGITS,"1E-3"]) ; }
	|!	PPM			{ #expr = #([DIGITS,"1E-6"]) ; }
	|!	PPB			{ #expr = #([DIGITS,"1E-9"]) ; }
	|!	PPTR			{ #expr = #([DIGITS,"1E-12"]) ; }
	|!	MOLE			{ #expr = #([DIGITS,"6.0221367E+23"]) ; }
	|!	STREADIAN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [RADIAN,"rad"], [RADIAN,"rad"])) ; }
	|!	HERZ			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [SECOND,"s"])) ; }
	|!	NEWTON			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], #([MULTIPLY,"."], [DIGITS,"1000"], [GRAM,"g"]), [MILLI_OR_METER,"m"]), #([MULTIPLY,"."], [SECOND,"s"], [SECOND,"s"]))) ; }
	|!	PASCAL			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [NEWTON,"N"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	JOULE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [NEWTON,"N"], [MILLI_OR_METER,"m"])) ; }
	|!	WATT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [JOULE,"J"], [SECOND,"s"])) ; }
	|!	AMPERE			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [COULOMB,"C"], [SECOND,"s"])) ; }
	|!	VOLT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [JOULE,"J"], [COULOMB,"C"])) ; }
	|!	FARAD			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [COULOMB,"C"], [VOLT,"V"])) ; }
	|!	OHM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [VOLT,"V"], [AMPERE,"A"])) ; }
	|!	SIEMENS			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [OHM,"Ohm"])) ; }
	|!	WEBER			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [VOLT,"V"], [SECOND,"s"])) ; }
	|!	TERA_OR_TESLA		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [WEBER,"Wb"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	HENRY			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [WEBER,"Wb"], [AMPERE,"A"])) ; }
	|!	LUMEN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [CANDELA,"cd"], [STREADIAN,"sd"])) ; }
	|!	LUX			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [LUMEN,"lm"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	BECQUEREL		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [SECOND,"s"])) ; }
	|!	GRAY			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [JOULE,"J"], #([MULTIPLY,"."], [DIGITS,"1000"], [GRAM,"g"]))) ; }
	|!	SIEVERT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [JOULE,"J"], #([MULTIPLY,"."], [DIGITS,"1000"], [GRAM,"g"]))) ; }
	|!	GON			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"0.9"], [DEGREE_ANGLE,"deg"])) ; }
	|!	DEGREE_ANGLE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2"], #([MULTIPLY,"."], [PI,"[pi]"], #([DIVIDE,"/"], [RADIAN,"rad"], [DIGITS,"360"])))) ; }
	|!	MINUTE_ANGLE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DEGREE_ANGLE,"deg"], [DIGITS,"60"])) ; }
	|!	SECOND_ANGLE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [MINUTE_ANGLE,"'"], [DIGITS,"60"])) ; }
	|!	LITER			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"0.001"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])))) ; }
	|!	LITERCAP		{ #expr = #[LITER,"l"] ; }
	|!	ARE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"100"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	MINUTE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"60"], [SECOND,"s"])) ; }
	|!	HECTO_OR_HOUR		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"60"], [MINUTE,"min"])) ; }
	|!	DECI_OR_DAY		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"24"], [HECTO_OR_HOUR,"h"])) ; }
	|!	YEAR_TROPICAL		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"365.24219"], [DECI_OR_DAY,"d"])) ; }
	|!	YEAR_JULIAN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"365.25"], [DECI_OR_DAY,"d"])) ; }
	|!	YEAR_GREGORIAN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"365.2425"], [DECI_OR_DAY,"d"])) ; }
	|!	ATTO_OR_YEAR		{ #expr = #[YEAR_JULIAN,"a_j"] ; }
	|!	WEEK			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"7"], [DECI_OR_DAY,"d"])) ; }
	|!	MONTH_SYNODAL		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"29.53059"], [DECI_OR_DAY,"d"])) ; }
	|!	MONTH_JULIAN		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [YEAR_JULIAN,"a_j"], [DIGITS,"12"])) ; }
	|!	MONTH_GREGORIAN		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [YEAR_GREGORIAN,"a_g"], [DIGITS,"12"])) ; }
	|!	MONTH			{ #expr = #[MONTH_JULIAN,"m_j"] ; }
	|!	TONNE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E+6"], [GRAM,"g"])) ; }
	|!	BAR			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E+5"], [PASCAL,"Pa"])) ; }
	|!	MICRO_OR_UAMU		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.6605402E-24"], [GRAM,"g"])) ; }
	|!	ELECTRON_VOLT		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [ELEMENTARY_CHARGE,"[e]"], [VOLT,"V"])) ; }
	|!	ASTRO_UNIT		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"149597.870E+6"], [MILLI_OR_METER,"m"])) ; }
	|!	PARSEC			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3.085678E+16"], [MILLI_OR_METER,"m"])) ; }
	|!	VELOCITY_OF_LIGHT	{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"299792458"], #([DIVIDE,"/"], [MILLI_OR_METER,"m"], [SECOND,"s"]))) ; }
	|!	PLANCK			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6.6260755E-34"], #([DIVIDE,"/"], [JOULE,"J"], [SECOND,"s"]))) ; }
	|!	BOLTZMAN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.380658E-23"], #([DIVIDE,"/"], [JOULE,"J"], [KELVIN,"K"]))) ; }
	|!	PERMITTIVITY_VACUUM	{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"8.854187817E-12"], #([DIVIDE,"/"], [FARAD,"F"], [MILLI_OR_METER,"m"]))) ; }
	|!	PERMEABILITY_VACUUM	{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], #([MULTIPLY,"."], [PI,"[pi]"], [DIGITS,"4E-7"]), [NEWTON,"N"]), #([MULTIPLY,"."], [AMPERE,"A"], [AMPERE,"A"]))) ; }
	|!	ELEMENTARY_CHARGE	{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.60217733E-19"], [COULOMB,"C"])) ; }
	|!	ELECTRON_MASS		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"9.1093897E-28"], [GRAM,"g"])) ; }
	|!	PROTON_MASS		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.6726231E-24"], [GRAM,"g"])) ; }
	|!	GRAVITY			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6.67259E-11"], #([DIVIDE,"/"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])), #([MULTIPLY,"."], #([MULTIPLY,"."], [DIGITS,"1000"], [GRAM,"g"]), #([MULTIPLY,"."], [SECOND,"s"], [SECOND,"s"]))))) ; }
	|!	ACCELERATION		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"9.80665"], #([DIVIDE,"/"], [MILLI_OR_METER,"m"], #([MULTIPLY,"."], [SECOND,"s"], [SECOND,"s"])))) ; }
	|!	ATOMOSPHERE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"101325"], [PASCAL,"Pa"])) ; }
	|!	LIGHT_YEAR		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [VELOCITY_OF_LIGHT,"[c]"], [YEAR_JULIAN,"a_j"])) ; }
	|!	GRAM_FORCE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [GRAM,"g"], [ACCELERATION,"[g]"])) ; }
	|!	POUND_FORCE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [POUND,"lb_av"], [ACCELERATION,"[g]"])) ; }
	|!	KAYSER			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"100"], [MILLI_OR_METER,"m"])) ; }
	|!	GAL			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,".01"], [MILLI_OR_METER,"m"]), #([MULTIPLY,"."], [SECOND,"s"], [SECOND,"s"]))) ; }
	|!	DYNE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [GRAM,"g"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,".01"], [MILLI_OR_METER,"m"]), #([MULTIPLY,"."], [SECOND,"s"], [SECOND,"s"])))) ; }
	|!	ERG			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DYNE,"dyn"], #([MULTIPLY,"."], [DIGITS,".01"], [MILLI_OR_METER,"m"]))) ; }
	|!	PETA_OR_POISE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DYNE,"dyn"], [SECOND,"s"]), #([MULTIPLY,"."], [DIGITS,".0001"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])))) ; }
	|!	BIOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"10"], [AMPERE,"A"])) ; }
	|!	STOKES			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,".0001"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])), [SECOND,"s"])) ; }
	|!	MAXWELL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-8"], [WEBER,"Wb"])) ; }
	|!	GIGA_OR_GAUSS		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-4"], [TERA_OR_TESLA,"T"])) ; }
	|!	OERSTED			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], #([DIVIDE,"/"], [DIGITS,"250"], [PI,"[pi]"]), [AMPERE,"A"]), [MILLI_OR_METER,"m"])) ; }
	|!	GILBERT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [OERSTED,"Oe"], #([MULTIPLY,"."], [DIGITS,".01"], [MILLI_OR_METER,"m"]))) ; }
	|!	STILB			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [CANDELA,"cd"], #([MULTIPLY,"."], [DIGITS,".0001"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])))) ; }
	|!	LAMBERT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([DIVIDE,"/"], [CANDELA,"cd"], #([MULTIPLY,"."], [DIGITS,".0001"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))), [PI,"[pi]"])) ; }
	|!	PHOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-4"], [LUX,"lx"])) ; }
	|!	CURIE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3.7E+10"], [BECQUEREL,"Bq"])) ; }
	|!	ROENTGEN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2.58E-4"], #([DIVIDE,"/"], [COULOMB,"C"], #([MULTIPLY,"."], [DIGITS,"1000"], [GRAM,"g"])))) ; }
	|!	RAD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"100"], #([DIVIDE,"/"], [ERG,"erg"], [GRAM,"g"]))) ; }
	|!	REM			{ #expr = #[RAD,"RAD"] ; }
	|!	INCH			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2.54E-2"], [MILLI_OR_METER,"m"])) ; }
	|!	FOOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [INCH,"[in_i]"])) ; }
	|!	YARD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3"], [FOOT,"[ft_i]"])) ; }
	|!	SMILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"5280"], [FOOT,"[ft_i]"])) ; }
	|!	FATHOM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6"], [FOOT,"[ft_i]"])) ; }
	|!	NMILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1852"], [MILLI_OR_METER,"m"])) ; }
	|!	KNOT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [NMILE,"[nmi_i]"], [HECTO_OR_HOUR,"h"])) ; }
	|!	SQ_INCH			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [INCH,"[in_i]"], [INCH,"[in_i]"])) ; }
	|!	SQ_FOOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [FOOT,"[ft_i]"], [FOOT,"[ft_i]"])) ; }
	|!	SQ_YARD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [YARD,"[yd_i]"], [YARD,"[yd_i]"])) ; }
	|!	CU_INCH			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [INCH,"[in_i]"], #([MULTIPLY,"."], [INCH,"[in_i]"], [INCH,"[in_i]"]))) ; }
	|!	CU_FOOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [FOOT,"[ft_i]"], #([MULTIPLY,"."], [FOOT,"[ft_i]"], [FOOT,"[ft_i]"]))) ; }
	|!	CU_YARD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [YARD,"[yd_i]"], #([MULTIPLY,"."], [YARD,"[yd_i]"], [YARD,"[yd_i]"]))) ; }
	|!	BOARD_FOOT		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"144"], [CU_INCH,"[cin_i]"])) ; }
	|!	CORD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"128"], [CU_FOOT,"[cft_i]"])) ; }
	|!	MIL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-3"], [INCH,"[in_i]"])) ; }
	|!	CIRC_MIL		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #([DIVIDE,"/"], [PI,"[pi]"], [DIGITS,"4"]),  #([MULTIPLY,"."], [MIL,"[mil_i]"], [MIL,"[mil_i]"]))) ; }
	|!	HAND			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4"], [INCH,"[in_i]"])) ; }
	|!	UFOOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1200"], #([DIVIDE,"/"], [MILLI_OR_METER,"m"], [DIGITS,"3937"]))) ; }
	|!	UYARD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3"], [UFOOT,"[ft_us]"])) ; }
	|!	UINCH			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UFOOT,"[ft_us]"], [DIGITS,"12"])) ; }
	|!	UROD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"16.5"], [UFOOT,"[ft_us]"])) ; }
	|!	GCHAIN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4"], [UROD,"[rd_us]"])) ; }
	|!	GLINK			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [GCHAIN,"[ch_us]"], [DIGITS,"100"])) ; }
	|!	RCHAIN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"100"], [UFOOT,"[ft_us]"])) ; }
	|!	RLINK			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [RCHAIN,"[rch_us]"], [DIGITS,"100"])) ; }
	|!	UFATHOM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6"], [UFOOT,"[ft_us]"])) ; }
	|!	FURLONG			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"40"], [UROD,"[rd_us]"])) ; }
	|!	UMILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"8"], [FURLONG,"[fur_us]"])) ; }
	|!	UACRE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"160"], #([MULTIPLY,"."], [UROD,"[rd_us]"], [UROD,"[rd_us]"]))) ; }
	|!	SQ_ROD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [UROD,"[rd_us]"], [UROD,"[rd_us]"])) ; }
	|!	SQ_MILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [UMILE,"[mi_us]"], [UMILE,"[mi_us]"])) ; }
	|!	SECTION			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [UMILE,"[mi_us]"], [UMILE,"[mi_us]"])) ; }
	|!	TOWNSHIP		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"36"], [SECTION,"[sct]"])) ; }
	|!	UMIL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-3"], [INCH,"[in_us]"])) ; }
	|!	BINCH			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2.539998E-2"], [MILLI_OR_METER,"m"])) ; }
	|!	BFOOT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [BINCH,"[in_br]"])) ; }
	|!	BROD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"16.5"], [BFOOT,"[ft_br]"])) ; }
	|!	BCHAIN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4"], [BROD,"[rd_br]"])) ; }
	|!	BLINK			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BCHAIN,"[ch_br]"], [DIGITS,"100"])) ; }
	|!	BFATHOM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6"], [BFOOT,"[ft_br]"])) ; }
	|!	PACE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2.5"], [BFOOT,"[ft_br]"])) ; }
	|!	BYARD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3"], [BFOOT,"[ft_br]"])) ; }
	|!	BMILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"5280"], [BFOOT,"[ft_br]"])) ; }
	|!	BNMILE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"6080"], [BFOOT,"[ft_br]"])) ; }
	|!	BKNOT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BNMILE,"[nmi_br]"], [HECTO_OR_HOUR,"h"])) ; }
	|!	BACRE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4840"], #([MULTIPLY,"."], [BYARD,"[yd_br]"], [BYARD,"[yd_br]"]))) ; }
	|!	UGALLON			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"231"], [CU_INCH,"[cin_i]"])) ; }
	|!	UBARREL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"42"], [UGALLON,"[gal_us]"])) ; }
	|!	UQUART			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UGALLON,"[gal_us]"], [DIGITS,"4"])) ; }
	|!	UPINT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UQUART,"[qt_us]"], [DIGITS,"2"])) ; }
	|!	UGILL			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UPINT,"[pt_us]"], [DIGITS,"4"])) ; }
	|!	UFLOUNCE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UGILL,"[gil_us]"], [DIGITS,"4"])) ; }
	|!	UFLDRAM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UFLOUNCE,"[foz_us]"], [DIGITS,"8"])) ; }
	|!	UMINIM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UFLDRAM,"[fdr_us]"], [DIGITS,"60"])) ; }
	|!	UCORD			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"128"], [CU_FOOT,"[cft_i]"])) ; }
	|!	UBUSHEL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2150.42"], [CU_INCH,"[cin_i]"])) ; }
	|!	WINGALLON		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UBUSHEL,"[bu_us]"], [DIGITS,"8"])) ; }
	|!	UPECK			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UBUSHEL,"[bu_us]"], [DIGITS,"4"])) ; }
	|!	UDRYQUART		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UPECK,"[pk_us]"], [DIGITS,"8"])) ; }
	|!	UDRYPINT		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UDRYQUART,"[dqt_us]"], [DIGITS,"2"])) ; }
	|!	UTABLESPOON		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UFLOUNCE,"[foz_us]"], [DIGITS,"2"])) ; }
	|!	UTEASPOON		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [UTABLESPOON,"[tbs_us]"], [DIGITS,"3"])) ; }
	|!	UCUP			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"16"], [UTABLESPOON,"[tbs_us]"])) ; }
	|!	BGALLON			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.54609"], [LITER,"l"])) ; }
	|!	BPECK			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2"], [BGALLON,"gal_br"])) ; }
	|!	BBUSHEL			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4"], [BPECK,"pk_br"])) ; }
	|!	BQUART			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BGALLON,"[gal_br]"], [DIGITS,"4"])) ; }
	|!	BPINT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BQUART,"[qt_br]"], [DIGITS,"2"])) ; }
	|!	BGILL			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BPINT,"[pt_br]"], [DIGITS,"4"])) ; }
	|!	BFLOUNCE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BGILL,"[gil_br]"], [DIGITS,"5"])) ; }
	|!	BFLDRAM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BFLOUNCE,"[foz_br]"], [DIGITS,"8"])) ; }
	|!	BMINIM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [BFLDRAM,"[fdr_br]"], [DIGITS,"60"])) ; }
	|!	GRAIN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"64.79891E-3"], [GRAM,"g"])) ; }
	|!	POUND			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"7000"], [GRAIN,"[gr]"])) ; }
	|!	OUNCE			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [POUND,"[lb_av]"], [DIGITS,"16"])) ; }
	|!	DRAM			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [OUNCE,"[oz_av]"], [DIGITS,"16"])) ; }
	|!	SCWT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"100"], [POUND,"[lb_av]"])) ; }
	|!	LCWT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"112"], [POUND,"[lb_av]"])) ; }
	|!	STON			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"20"], [SCWT,"[scwt_av]"])) ; }
	|!	LTON			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"20"], [LCWT,"[lcwt_av]"])) ; }
	|!	STONE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"14"], [POUND,"[lb_av]"])) ; }
	|!	PENNYWEIGHT		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"24"], [GRAIN,"[gr]"])) ; }
	|!	TROYOUNCE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"20"], [PENNYWEIGHT,"[pwt_tr]"])) ; }
	|!	TROYPOUND		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [TROYOUNCE,"[oz_tr]"])) ; }
	|!	SCRUPLE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"20"], [GRAIN,"[gr]"])) ; }
	|!	APDRAM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"3"], [SCRUPLE,"[sc_ap]"])) ; }
	|!	APOUNCE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"8"], [APDRAM,"[dr_ap]"])) ; }
	|!	APPOUND			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [APOUNCE,"[oz_ap]"])) ; }
	|!	LINE			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [INCH,"[in_i]"], [DIGITS,"12"])) ; }
	|!	POINT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [LINE,"[lne]"], [DIGITS,"6"])) ; }
	|!	PICA			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [POINT,"[pnt]"])) ; }
	|!	PPOINT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"0.013837"], [INCH,"[in_i]"])) ; }
	|!	PPICA			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [PPOINT,"[pnt_pr]"])) ; }
	|!	PIED			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"32.48E-2"], [MILLI_OR_METER,"m"])) ; }
	|!	POUCE			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [PIED,"[pied]"], [DIGITS,"12"])) ; }
	|!	LIGNE			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [POUCE,"[pouce]"], [DIGITS,"12"])) ; }
	|!	DIDOT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [LIGNE,"[ligne]"], [DIGITS,"6"])) ; }
	|!	CICERO			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"12"], [DIDOT,"[didot]"])) ; }
	|!	CALORIE15		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.18580"], [JOULE,"J"])) ; }
	|!	CALORIE20		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.18190"], [JOULE,"J"])) ; }
	|!	CALORIEM		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.19002"], [JOULE,"J"])) ; }
	|!	CALORIEIT		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.1868"], [JOULE,"J"])) ; }
	|!	CALORIETH		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4.184"], [JOULE,"J"])) ; }
	|!	CALORIE			{ #expr = #[CALORIETH,"cal_th"] ; }
	|!	CALORIEN		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1000"], [CALORIETH,"cal_th"])) ; }
	|!	BTU39			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.05967E+3"], [JOULE,"J"])) ; }
	|!	BTU59			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.05480E+3"], [JOULE,"J"])) ; }
	|!	BTU60			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.05468E+3"], [JOULE,"J"])) ; }
	|!	BTUM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.05587E+3"], [JOULE,"J"])) ; }
	|!	BTUIT			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.05505585262E+3"], [JOULE,"J"])) ; }
	|!	BTUTH			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1.054350E+3"], [JOULE,"J"])) ; }
	|!	BTU			{ #expr = #[BTUTH,"Btu_th"] ; }
	|!	METERH2O		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"9.80665E+3"], [PASCAL,"Pa"])) ; }
	|!	METERMERC		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"133.3220E+3"], [PASCAL,"Pa"])) ; }
	|!	INCHH2O			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [METERH2O,"m[H2O]"], [INCH,"[in_i]"]), [MILLI_OR_METER,"m"])) ; }
	|!	INCHMERC		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [METERMERC,"m[Hg]"], [INCH,"[in_i]"]), [MILLI_OR_METER,"m"])) ; }
	|!	PERVASCRES		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [METERMERC,"m[Hg]"], [SECOND,"s"]), [LITER,"l"])) ; }
	|!	DIOPTER			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [MILLI_OR_METER,"m"])) ; }
	|!	MESH			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [INCH,"[in_i]"])) ; }
	|!	CHARRIERE		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,"1E-3"], [MILLI_OR_METER,"m"]), [PI,"[pi]"])) ; }
	|!	DROP			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,"1E-3"], [LITER,"l"]), [DIGITS,"12"])) ; }
	|!	EQUIVALENTS		{ #expr = #[MOLE,"mol"] ; }
	|!	OSMOLE			{ #expr = #[MOLE,"mol"] ; }
	|!	GRAMPERCENT		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [GRAM,"g"], #([MULTIPLY,"."], [DIGITS,"0.1"], [LITER,"l"]))) ; }
	|!	SVEDBERG		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-13"], [SECOND,"s"])) ; }
	|!	LOWPOWERFIELD		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"100"], [HIGHPOWERFIELD,"[HPF]"])) ; }
	|!	KATAL			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [MOLE,"mol"], [SECOND,"s"])) ; }
	|!	UNIT			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [DIGITS,"1E-6"], [MOLE,"mol"]), [MINUTE,"min"])) ; }
	|!	STERE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	ANGSTROM		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-10"], [MILLI_OR_METER,"m"])) ; }
	|!	BARN			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-28"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"]))) ; }
	|!	ATMOSPHERE		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E+7"], #([DIVIDE,"/"], [GRAM_FORCE,"gf"], #([MULTIPLY,"."], [MILLI_OR_METER,"m"], [MILLI_OR_METER,"m"])))) ; }
	|!	PSI			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [POUND_FORCE,"lbf_av"], #([MULTIPLY,"."], [INCH,"in_i"], [INCH,"in_i"]))) ; }
	|!	CIRCLE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"2"], #([MULTIPLY,"."], [PI,"[pi]"], [RADIAN,"rad"]))) ; }
	|!	SPHERE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"4"], #([MULTIPLY,"."], [PI,"[pi]"], [STREADIAN,"sr"]))) ; }
	|!	CARATM			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"0.2"], [GRAM,"g"])) ; }
	|!	CARATG			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [DIGITS,"24"])) ; }
	|!	BYTE			{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"8"], [BIT,"bit"])) ; }
	|!	BAUD			{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], [DIGITS,"1"], [SECOND,"s"])) ; }
	|!	HORSEPOWER		{ #expr = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"], #([MULTIPLY,"."], [FOOT,"[ft_i]"], [POUND_FORCE,"lbf_av"]), [SECOND,"s"])) ; }
	|!	METABOLICEQUIV		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], [DIGITS,"1E-6"], #([DIVIDE,"/"], #([DIVIDE,"/"], [LITER,"l"], [MINUTE,"min"]), [GRAM,"g"]))) ; }
	|!	MHO			{ #expr = #[SIEMENS,"S"] ; }
	|!	CELCIUSDELTA		{ #expr = #[KELVINDELTA,"degK_delta"] ; }
	|!	FAHRENDELTA		{ #expr = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #[KELVINDELTA,"degK_delta"], #([DIVIDE,"/"], [DIGITS,"5"], [DIGITS,"9"]))) ; }
	|!	ANNOTATION		{ #expr = #[DIGITS,"1"] ; }
	|	.
	;

class UCUMParser extends Parser;

options {
        buildAST=true;
	k=2; 		// Lookahead size
}

tokens {
	PREFIX_ATOM="prefix";
	EVAL_ATOM="eval";
}

{
	int debugMode = 0;

	void setDebugMode(int debugMode)	{ this.debugMode = debugMode; }

	class OurNodeCloner {
		AST clone(AST x) {
			AST xCopy = astFactory.dup(x);
			AST child = x.getFirstChild();
			while (child != null) {
				if (debugMode > 2) { System.err.println("value: duplicating child of x = "+child.toStringList()); }
				xCopy.addChild(clone(child));
				child = child.getNextSibling();
			}
			return xCopy;
		}
	};
	OurNodeCloner ourNodeCloner = new OurNodeCloner();
}

factor:		DIGITS
	;

exponent:!	s:SIGN d:DIGITS		{ String value = s.getText()+d.getText(); #exponent = #([DIGITS,value]); }
	|	DIGITS
	;

atom:		MILLI_OR_METER (ANNOTATION)?
	|	GRAM (ANNOTATION)?
	|	SECOND (ANNOTATION)?
	|	RADIAN (ANNOTATION)?
	|	KELVIN (ANNOTATION)?
	|	COULOMB (ANNOTATION)?
	|	CANDELA (ANNOTATION)?
	|	TEN (ANNOTATION)?
	|	TENALT (ANNOTATION)?
	|	PI (ANNOTATION)?
	|	PERCENT (ANNOTATION)?
	|	PPTH (ANNOTATION)?
	|	PPM (ANNOTATION)?
	|	PPB (ANNOTATION)?
	|	PPTR (ANNOTATION)?
	|	MOLE (ANNOTATION)?
	|	STREADIAN (ANNOTATION)?
	|	HERZ (ANNOTATION)?
	|	NEWTON (ANNOTATION)?
	|	PASCAL (ANNOTATION)?
	|	JOULE (ANNOTATION)?
	|	WATT (ANNOTATION)?
	|	AMPERE (ANNOTATION)?
	|	VOLT (ANNOTATION)?
	|	FARAD (ANNOTATION)?
	|	OHM (ANNOTATION)?
	|	SIEMENS (ANNOTATION)?
	|	WEBER (ANNOTATION)?
	|	CELSIUS (ANNOTATION)?
	|	TERA_OR_TESLA (ANNOTATION)?
	|	HENRY (ANNOTATION)?
	|	LUMEN (ANNOTATION)?
	|	LUX (ANNOTATION)?
	|	BECQUEREL (ANNOTATION)?
	|	GRAY (ANNOTATION)?
	|	SIEVERT (ANNOTATION)?
	|	GON (ANNOTATION)?
	|	DEGREE_ANGLE (ANNOTATION)?
	|	MINUTE_ANGLE (ANNOTATION)?
	|	SECOND_ANGLE (ANNOTATION)?
	|	LITER (ANNOTATION)?
	|	LITERCAP (ANNOTATION)?
	|	ARE (ANNOTATION)?
	|	MINUTE (ANNOTATION)?
	|	HECTO_OR_HOUR (ANNOTATION)?
	|	DECI_OR_DAY (ANNOTATION)?
	|	YEAR_TROPICAL (ANNOTATION)?
	|	YEAR_JULIAN (ANNOTATION)?
	|	YEAR_GREGORIAN (ANNOTATION)?
	|	ATTO_OR_YEAR (ANNOTATION)?
	|	WEEK (ANNOTATION)?
	|	MONTH_SYNODAL (ANNOTATION)?
	|	MONTH_JULIAN (ANNOTATION)?
	|	MONTH_GREGORIAN (ANNOTATION)?
	|	MONTH (ANNOTATION)?
	|	TONNE (ANNOTATION)?
	|	BAR (ANNOTATION)?
	|	MICRO_OR_UAMU (ANNOTATION)?
	|	ELECTRON_VOLT (ANNOTATION)?
	|	ASTRO_UNIT (ANNOTATION)?
	|	PARSEC (ANNOTATION)?
	|	VELOCITY_OF_LIGHT (ANNOTATION)?
	|	PLANCK (ANNOTATION)?
	|	BOLTZMAN (ANNOTATION)?
	|	PERMITTIVITY_VACUUM (ANNOTATION)?
	|	PERMEABILITY_VACUUM (ANNOTATION)?
	|	ELEMENTARY_CHARGE (ANNOTATION)?
	|	ELECTRON_MASS (ANNOTATION)?
	|	PROTON_MASS (ANNOTATION)?
	|	GRAVITY (ANNOTATION)?
	|	ACCELERATION (ANNOTATION)?
	|	ATOMOSPHERE (ANNOTATION)?
	|	LIGHT_YEAR (ANNOTATION)?
	|	GRAM_FORCE (ANNOTATION)?
	|	POUND_FORCE (ANNOTATION)?
	|	KAYSER (ANNOTATION)?
	|	GAL (ANNOTATION)?
	|	DYNE (ANNOTATION)?
	|	ERG (ANNOTATION)?
	|	PETA_OR_POISE (ANNOTATION)?
	|	BIOT (ANNOTATION)?
	|	STOKES (ANNOTATION)?
	|	MAXWELL (ANNOTATION)?
	|	GIGA_OR_GAUSS (ANNOTATION)?
	|	OERSTED (ANNOTATION)?
	|	GILBERT (ANNOTATION)?
	|	STILB (ANNOTATION)?
	|	LAMBERT (ANNOTATION)?
	|	PHOT (ANNOTATION)?
	|	CURIE (ANNOTATION)?
	|	ROENTGEN (ANNOTATION)?
	|	RAD (ANNOTATION)?
	|	REM (ANNOTATION)?
	|	INCH (ANNOTATION)?
	|	FOOT (ANNOTATION)?
	|	YARD (ANNOTATION)?
	|	SMILE (ANNOTATION)?
	|	FATHOM (ANNOTATION)?
	|	NMILE (ANNOTATION)?
	|	KNOT (ANNOTATION)?
	|	SQ_INCH (ANNOTATION)?
	|	SQ_FOOT (ANNOTATION)?
	|	SQ_YARD (ANNOTATION)?
	|	CU_INCH (ANNOTATION)?
	|	CU_FOOT (ANNOTATION)?
	|	CU_YARD (ANNOTATION)?
	|	BOARD_FOOT (ANNOTATION)?
	|	CORD (ANNOTATION)?
	|	MIL (ANNOTATION)?
	|	CIRC_MIL (ANNOTATION)?
	|	HAND (ANNOTATION)?
	|	UFOOT (ANNOTATION)?
	|	UYARD (ANNOTATION)?
	|	UINCH (ANNOTATION)?
	|	UROD (ANNOTATION)?
	|	GCHAIN (ANNOTATION)?
	|	GLINK (ANNOTATION)?
	|	RCHAIN (ANNOTATION)?
	|	RLINK (ANNOTATION)?
	|	UFATHOM (ANNOTATION)?
	|	FURLONG (ANNOTATION)?
	|	UMILE (ANNOTATION)?
	|	UACRE (ANNOTATION)?
	|	SQ_ROD (ANNOTATION)?
	|	SQ_MILE (ANNOTATION)?
	|	SECTION (ANNOTATION)?
	|	TOWNSHIP (ANNOTATION)?
	|	UMIL (ANNOTATION)?
	|	BINCH (ANNOTATION)?
	|	BFOOT (ANNOTATION)?
	|	BROD (ANNOTATION)?
	|	BCHAIN (ANNOTATION)?
	|	BLINK (ANNOTATION)?
	|	BFATHOM (ANNOTATION)?
	|	PACE (ANNOTATION)?
	|	BYARD (ANNOTATION)?
	|	BMILE (ANNOTATION)?
	|	BNMILE (ANNOTATION)?
	|	BKNOT (ANNOTATION)?
	|	BACRE (ANNOTATION)?
	|	UGALLON (ANNOTATION)?
	|	UBARREL (ANNOTATION)?
	|	UQUART (ANNOTATION)?
	|	UPINT (ANNOTATION)?
	|	UGILL (ANNOTATION)?
	|	UFLOUNCE (ANNOTATION)?
	|	UFLDRAM (ANNOTATION)?
	|	UMINIM (ANNOTATION)?
	|	UCORD (ANNOTATION)?
	|	UBUSHEL (ANNOTATION)?
	|	WINGALLON (ANNOTATION)?
	|	UPECK (ANNOTATION)?
	|	UDRYQUART (ANNOTATION)?
	|	UDRYPINT (ANNOTATION)?
	|	UTABLESPOON (ANNOTATION)?
	|	UTEASPOON (ANNOTATION)?
	|	UCUP (ANNOTATION)?
	|	BGALLON (ANNOTATION)?
	|	BPECK (ANNOTATION)?
	|	BBUSHEL (ANNOTATION)?
	|	BQUART (ANNOTATION)?
	|	BPINT (ANNOTATION)?
	|	BGILL (ANNOTATION)?
	|	BFLOUNCE (ANNOTATION)?
	|	BFLDRAM (ANNOTATION)?
	|	BMINIM (ANNOTATION)?
	|	GRAIN (ANNOTATION)?
	|	POUND (ANNOTATION)?
	|	OUNCE (ANNOTATION)?
	|	DRAM (ANNOTATION)?
	|	SCWT (ANNOTATION)?
	|	LCWT (ANNOTATION)?
	|	STON (ANNOTATION)?
	|	LTON (ANNOTATION)?
	|	STONE (ANNOTATION)?
	|	PENNYWEIGHT (ANNOTATION)?
	|	TROYOUNCE (ANNOTATION)?
	|	TROYPOUND (ANNOTATION)?
	|	SCRUPLE (ANNOTATION)?
	|	APDRAM (ANNOTATION)?
	|	APOUNCE (ANNOTATION)?
	|	APPOUND (ANNOTATION)?
	|	LINE (ANNOTATION)?
	|	POINT (ANNOTATION)?
	|	PICA (ANNOTATION)?
	|	PPOINT (ANNOTATION)?
	|	PPICA (ANNOTATION)?
	|	PIED (ANNOTATION)?
	|	POUCE (ANNOTATION)?
	|	LIGNE (ANNOTATION)?
	|	DIDOT (ANNOTATION)?
	|	CICERO (ANNOTATION)?
	|	FAHRENHEIT (ANNOTATION)?
	|	CALORIE15 (ANNOTATION)?
	|	CALORIE20 (ANNOTATION)?
	|	CALORIEM (ANNOTATION)?
	|	CALORIEIT (ANNOTATION)?
	|	CALORIETH (ANNOTATION)?
	|	CALORIE (ANNOTATION)?
	|	CALORIEN (ANNOTATION)?
	|	BTU39 (ANNOTATION)?
	|	BTU59 (ANNOTATION)?
	|	BTU60 (ANNOTATION)?
	|	BTUM (ANNOTATION)?
	|	BTUIT (ANNOTATION)?
	|	BTUTH (ANNOTATION)?
	|	BTU (ANNOTATION)?
	|	METERH2O (ANNOTATION)?
	|	METERMERC (ANNOTATION)?
	|	INCHH2O (ANNOTATION)?
	|	INCHMERC (ANNOTATION)?
	|	PERVASCRES (ANNOTATION)?
	|	DIOPTER (ANNOTATION)?
	|	MESH (ANNOTATION)?
	|	CHARRIERE (ANNOTATION)?
	|	DROP (ANNOTATION)?
	|	HU (ANNOTATION)?
	|	EQUIVALENTS (ANNOTATION)?
	|	OSMOLE (ANNOTATION)?
	|	ACIDITY (ANNOTATION)?
	|	GRAMPERCENT (ANNOTATION)?
	|	SVEDBERG (ANNOTATION)?
	|	HIGHPOWERFIELD (ANNOTATION)?
	|	LOWPOWERFIELD (ANNOTATION)?
	|	KATAL (ANNOTATION)?
	|	UNIT (ANNOTATION)?
	|	INTUNIT (ANNOTATION)?
	|	ARBUNIT (ANNOTATION)?
	|	USPUNIT (ANNOTATION)?
	|	GPLUNIT (ANNOTATION)?
	|	MPLUNIT (ANNOTATION)?
	|	APLUNIT (ANNOTATION)?
	|	BETHESDAUNIT (ANNOTATION)?
	|	TODDUNIT (ANNOTATION)?
	|	DYEUNIT (ANNOTATION)?
	|	SOMOGYIUNIT (ANNOTATION)?
	|	BODANSKYUNIT (ANNOTATION)?
	|	KIMGARMUNIT (ANNOTATION)?
	|	KUNKELUNIT (ANNOTATION)?
	|	MACLAGANUNIT (ANNOTATION)?
	|	TUBERUNIT (ANNOTATION)?
	|	NEPER (ANNOTATION)?
	|	BEL (ANNOTATION)?
	|	BELSOUNDPRES (ANNOTATION)?
	|	BELVOLT (ANNOTATION)?
	|	BELMILLIVOLT (ANNOTATION)?
	|	BELMICROVOLT (ANNOTATION)?
	|	BELWATT (ANNOTATION)?
	|	BELKILOWATT (ANNOTATION)?
	|	STERE (ANNOTATION)?
	|	ANGSTROM (ANNOTATION)?
	|	BARN (ANNOTATION)?
	|	ATMOSPHERE (ANNOTATION)?
	|	PSI (ANNOTATION)?
	|	CIRCLE (ANNOTATION)?
	|	SPHERE (ANNOTATION)?
	|	CARATM (ANNOTATION)?
	|	CARATG (ANNOTATION)?
	|	BIT_S (ANNOTATION)?
	|	BIT (ANNOTATION)?
	|	BYTE (ANNOTATION)?
	|	BAUD (ANNOTATION)?
	|	HORSEPOWER (ANNOTATION)?
	|	METABOLICEQUIV (ANNOTATION)?
	|	HOMEOPOTDEC (ANNOTATION)?
	|	HOMEOPOTCEN (ANNOTATION)?
	|	MHO (ANNOTATION)?
	|	KELVINDELTA (ANNOTATION)?
	|	CELCIUSDELTA (ANNOTATION)?
	|	FAHRENDELTA (ANNOTATION)?
	;


simple:		(options { warnWhenFollowAmbig=false; }: atom)
	|	ANNOTATION
	|!	DEKA		adeka:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #adeka,   [DIGITS,"10"])) ; }
	|!	HECTO_OR_HOUR	ahecto:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #ahecto,  [DIGITS,"100"])) ; }
	|!	KILO		akilo:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #akilo,   [DIGITS,"1000"])) ; }
	|!	MEGA		amega:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #amega,   [DIGITS,"1E+6"])) ; }
	|!	GIGA_OR_GAUSS	agiga:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #agiga,   [DIGITS,"1E+9"])) ; }
	|!	TERA_OR_TESLA	atera:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #atera,   [DIGITS,"1E+12"])) ; }
	|!	PETA_OR_POISE	apeta:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #apeta,   [DIGITS,"1E+15"])) ; }
	|!	EXA		aexa:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #aexa,    [DIGITS,"1E+18"])) ; }
	|!	ZETTA		azetta:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #azetta,  [DIGITS,"1E+21"])) ; }
	|!	YOTTA		ayotta:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #ayotta,  [DIGITS,"1E+24"])) ; }
	|!	DECI_OR_DAY	adeci:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #adeci,   [DIGITS,"10"])) ; }
	|!	CENTI		acenti:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #acenti,  [DIGITS,"100"])) ; }
	|!	MILLI_OR_METER	amilli:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #amilli,  [DIGITS,"1000"])) ; }
	|!	MICRO_OR_UAMU	amicro:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #amicro,  [DIGITS,"1E+6"])) ; }
	|!	NANO		anano:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #anano,   [DIGITS,"1E+9"])) ; }
	|!	PICO		apico:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #apico,   [DIGITS,"1E+12"])) ; }
	|!	FEMPTO		afempto:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #afempto, [DIGITS,"1E+15"])) ; }
	|!	ATTO_OR_YEAR	aatto:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #aatto,   [DIGITS,"1E+18"])) ; }
	|!	ZEPTO		azepto:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #azepto,  [DIGITS,"1E+21"])) ; }
	|!	YOCTO		ayocto:atom	{ #simple = #([EVAL_ATOM,"eval"], #([DIVIDE,"/"],   #ayocto,  [DIGITS,"1E+24"])) ; }
	|!	KIBI		akibi:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #akibi,  [DIGITS,"1024"])) ; }
	|!	MEBI		amebi:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #amebi,  [DIGITS,"1048576"])) ; }
	|!	GIBI		agibi:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #agibi,  [DIGITS,"1073741824"])) ; }
	|!	TEBI		atebi:atom	{ #simple = #([EVAL_ATOM,"eval"], #([MULTIPLY,"."], #atebi,  [DIGITS,"1099511627776"])) ; }
	|!	LPAREN! t:term RPAREN!		{ #simple = #([EVAL_ATOM,"eval"], #t) ; }
	|	factor
	;

power:!		s:simple (e:exponent)?		{ if (#e != null) {
							if (debugMode > 1) { System.err.println("power: simple = "+#s.toStringList()); }
							AST our_return = null;
							String exponent_string = #e.getText();
							if (debugMode > 1) { System.err.println("power: exponent_string = "+ exponent_string); }
							if (exponent_string.startsWith("+")) {
								exponent_string=exponent_string.substring(1);	// Integer.parseInt() does not like leading '+'
							}
							int exponent = Integer.parseInt(exponent_string);
							if (debugMode > 1) { System.err.println("power: exponent = "+ exponent); }
							if (exponent <= 0) {
								our_return = #([DIGITS,"1"]);
								while (exponent++ < 0) {
									AST new_node = #([DIVIDE,"/"], our_return, astFactory.dupList(#s));
									our_return = new_node;
								}
							}
							else if (exponent > 0) {
								our_return = astFactory.dupList(#s);
								while (exponent-- > 1) {
									if (debugMode > 1) { System.err.println("power: looping = "+our_return.toStringList()); }
									AST new_node = #([MULTIPLY,"."], our_return, astFactory.dupList(#s));
									our_return = new_node;
								}
							}
							if (debugMode > 1) { System.err.println("power: our_return = "+our_return.toStringList()); }
							#power = #([EVAL_ATOM,"eval"], our_return);
						  }
						  else {
							#power = #s ;
						  }
						}
	;

dm_operator:	DIVIDE^
	|	MULTIPLY^
	;

value:!		p:power (o:dm_operator v:value)?	{ if (#o != null && #v != null) {
								// We want a left associative tree, but cannot use left recursion
								//
								// Observe that the parser encounters and executes the right most actions first
								//
								// Want to add (o p) not to v, but to the deepest left child of v in a string of dm_operators
								
								if (debugMode > 1) { System.err.println("value: make tree left not right associative"); }
								if (debugMode > 1) { System.err.println("value: p = "+#p.toStringList()); }
								if (debugMode > 1) { System.err.println("value: v = "+#v.toStringList()); }

								AST our_parent = null;
								AST our_child = #v;
								while (our_child != null) {
									if (debugMode > 1) { System.err.println("value: checking "+ our_child.toStringList()); }
									int token = our_child.getType();
									if (token == DIVIDE || token == MULTIPLY) {
										if (debugMode > 1) { System.err.println("value: descending"); }
										our_parent = our_child;
										our_child = our_child.getFirstChild();	// keep descending left most branch
									}
									else {
										if (debugMode > 1) { System.err.println("value: stopping"); }
										break;			// our_parent is divide/multiply parent (if any); our_child is left child
									}
								}
								
								AST our_return = null;
								if (our_parent != null && our_child != null) {
									AST our_sibling = our_child.getNextSibling();
									// copy everything to make sure it is detached so as not to assume anything about how the tree works
									// astFactory.dupList(x) does NOT remove siblings, so do it manually
									our_child = ourNodeCloner.clone(our_child);
									if (debugMode > 1) { System.err.println("value: cloned child = "+our_child.toStringList()); }
									if (our_sibling != null) {
										our_sibling = ourNodeCloner.clone(our_sibling);
										if (debugMode > 1) { System.err.println("value: cloned sibling = "+our_sibling.toStringList()); }
									}
									AST new_node = #(#o, #p, our_child);
									if (debugMode > 1) { System.err.println("value: new node = "+new_node.toStringList()); }
									our_parent.setFirstChild(new_node);
									our_parent.addChild(our_sibling);
									our_return = #v;		// top of tree
									if (debugMode > 1) { System.err.println("value: rewriting as "+our_return.toStringList()); }
								}
								else {
									our_return = #(#o, #p, #v);
									if (debugMode > 1) { System.err.println("value: not rewriting, leave as "+our_return.toStringList()); }
								}
										
								#value = our_return;
							  }
							  else {
								if (debugMode > 1) { System.err.println("value: copying p into value"); }
								#value = #p ;
							  }
							}
	;

term:!		o:DIVIDE v:value			{ #term =  #(#o, [DIGITS,"1"], #v); }
	|!	vm:value				{ #term =  #([MULTIPLY,"."], [DIGITS,"1"], #vm); }
	;

line:		term EOL!
	|!	EOL!					{ #line = #[DIGITS,"1"]; /* e.g., for annotation alone */ }
	;

class UCUMLexer extends Lexer;

options {
    k=8; 				// Lookahead size
    charVocabulary='\u0000'..'\u007F'; 	// allow ascii
}

{
	int debugMode = 0;

	void setDebugMode(int debugMode)	{ this.debugMode = debugMode; }
}

ANNOTATION:	'{' (options {greedy=false;} :.)* '}' ;

SIGN:		'+' | '-' ;
DIGITS:		('0'..'9')+ ;
LPAREN: 	'(' ;
RPAREN: 	')' ;
LBRACE: 	'{' ;
RBRACE: 	'}' ;
MULTIPLY:	'.' ;
DIVIDE:		'/' ;

YOTTA:		'Y' ;
ZETTA:		'Z' ;
EXA:		'E' ;
MEGA:		'M' ;
KILO:		'k' ;
DEKA:		"da" ;
CENTI:		'c' ;
NANO:		'n' ;
PICO:		'p' ;
FEMPTO:		'f' ;
ZEPTO:		'z' ;
YOCTO:		'y' ;

MILLI_OR_METER:	'm' ;
TERA_OR_TESLA:	'T' ;
HECTO_OR_HOUR:	'h' ;
DECI_OR_DAY:	'd' ;
ATTO_OR_YEAR:	'a' ;
MICRO_OR_UAMU:	'u' ;
PETA_OR_POISE:	"P" ;
GIGA_OR_GAUSS:	'G' ;

GRAM:		'g' ;
SECOND:		's' ;
RADIAN:		"rad" ;
KELVIN:		'K' ;
COULOMB:	'C' ;
CANDELA:	"cd" ;
TEN:		"10*" ;
TENALT:		"10^" ;
PI:		'[' 'p' 'i' ']' ;
PERCENT:	"%" ;
PPTH:		'[' 'p' 'p' 't' 'h' ']' ;
PPM:		'[' 'p' 'p' 'm' ']' ;
PPB:		'[' 'p' 'p' 'b' ']' ;
PPTR:		'[' 'p' 'p' 't' 'r' ']' ;
MOLE:		"mol" ;
STREADIAN:	"sr" ;
HERZ:		"Hz" ;
NEWTON:		"N" ;
PASCAL:		"Pa" ;
JOULE:		"J" ;
WATT:		"W" ;
AMPERE:		"A" ;
VOLT:		"V" ;
FARAD:		"F" ;
OHM:		"Ohm" ;
SIEMENS:	"S" ;
WEBER:		"Wb" ;
CELSIUS:	"Cel" ;
HENRY:		"H" ;
LUMEN:		"lm" ;
LUX:		"lx" ;
BECQUEREL:	"Bq" ;
GRAY:		"Gy" ;
SIEVERT:	"Sv" ;
GON:		"gon" ;
DEGREE_ANGLE:	"deg" ;
MINUTE_ANGLE:	'\'' ;
SECOND_ANGLE:	'\'' '\'' ;
LITER:		"l" ;
LITERCAP:	"L" ;
ARE:		"ar" ;
MINUTE:		"min" ;
YEAR_TROPICAL:	"a_t" ;
YEAR_JULIAN:	"a_j" ;
YEAR_GREGORIAN:	"a_g" ;
WEEK:		"wk" ;
MONTH_SYNODAL:	"mo_s" ;
MONTH_JULIAN:	"mo_j" ;
MONTH_GREGORIAN:"mo_g" ;
MONTH:		"mo" ;
TONNE:		"t" ;
BAR:		"bar" ;
ELECTRON_VOLT:	"eV" ;
ASTRO_UNIT:	"AU" ;
PARSEC:		"pc" ;
VELOCITY_OF_LIGHT:	'[' 'c' ']' ;
PLANCK:			'[' 'h' ']' ;
BOLTZMAN:		'[' 'k' ']' ;
PERMITTIVITY_VACUUM:	'[' 'e' 'p' 's' '_' '0' ']' ;
PERMEABILITY_VACUUM:	'[' 'm' 'u' '_' '0' ']' ;
ELEMENTARY_CHARGE:	'[' 'e' ']' ;
ELECTRON_MASS:		'[' 'm' '_' 'e' ']' ;
PROTON_MASS:		'[' 'm' '_' 'p' ']' ;
GRAVITY:		'[' 'G' ']' ;
ACCELERATION:		'[' 'g' ']' ;
ATOMOSPHERE:		"atm" ;
LIGHT_YEAR:		'[' 'l' 'y' ']' ;
GRAM_FORCE:		"gf" ;
POUND_FORCE:		'[' 'l' 'b' 'f' '_' 'a' 'v' ']' ;
KAYSER:		"Ky" ;
GAL:		"Gal" ;
DYNE:		"dyn" ;
ERG:		"erg" ;
BIOT:		"Bi" ;
STOKES:		"St" ;
MAXWELL:	"Mx" ;
OERSTED:	"Oe" ;
GILBERT:	"Gb" ;
STILB:		"sb" ;
LAMBERT:	"Lmb" ;
PHOT:		"ph" ;
CURIE:		"Ci" ;
ROENTGEN:	"R" ;
RAD:		"RAD" ;
REM:		"REM" ;
INCH:		'[' 'i' 'n' '_' 'i' ']' ;
FOOT:		'[' 'f' 't' '_' 'i' ']' ;
YARD:		'[' 'y' 'd' '_' 'i' ']' ;
SMILE:		'[' 'm' 'i' '_' 'i' ']' ;
FATHOM:		'[' 'f' 't' 'h' '_' 'i' ']' ;
NMILE:		'[' 'n' 'm' 'i' '_' 'i' ']' ;
KNOT:		'[' 'k' 'n' '_' 'i' ']' ;
SQ_INCH:	'[' 's' 'i' 'n' '_' 'i' ']' ;
SQ_FOOT:	'[' 's' 'f' 't' '_' 'i' ']' ;
SQ_YARD:	'[' 's' 'y' 'd' '_' 'i' ']' ;
CU_INCH:	'[' 'c' 'i' 'n' '_' 'i' ']' ;
CU_FOOT:	'[' 'c' 'f' 't' '_' 'i' ']' ;
CU_YARD:	'[' 'c' 'y' 'd' '_' 'i' ']' ;
BOARD_FOOT:	'[' 'b' 'f' '_' 'i' ']' ;
CORD:		'[' 'c' 'r' '_' 'i' ']' ;
MIL:		'[' 'm' 'i' 'l' '_' 'i' ']' ;
CIRC_MIL:	'[' 'c' 'm' 'l' '_' 'i' ']' ;
HAND:		'[' 'h' 'd' '_' 'i' ']' ;
UFOOT:		'[' 'f' 't' '_' 'u' 's' ']' ;
UYARD:		'[' 'y' 'd' '_' 'u' 's' ']' ;
UINCH:		'[' 'i' 'n' '_' 'u' 's' ']' ;
UROD:		'[' 'r' 'd' '_' 'u' 's' ']' ;
GCHAIN:		'[' 'c' 'h' '_' 'u' 's' ']' ;
GLINK:		'[' 'l' 'k' '_' 'u' 's' ']' ;
RCHAIN:		'[' 'r' 'c' 'h' '_' 'u' 's' ']' ;
RLINK:		'[' 'r' 'l' 'k' '_' 'u' 's' ']' ;
UFATHOM:	'[' 'f' 't' 'h' '_' 'u' 's' ']' ;
FURLONG:	'[' 'f' 'u' 'r' '_' 'u' 's' ']' ;
UMILE:		'[' 'm' 'i' '_' 'u' 's' ']' ;
UACRE:		'[' 'a' 'c' 'r' '_' 'u' 's' ']' ;
SQ_ROD:		'[' 's' 'r' 'd' '_' 'u' 's' ']' ;
SQ_MILE:	'[' 's' 'm' 'i' '_' 'u' 's' ']' ;
SECTION:	'[' 's' 'c' 't' ']' ;
TOWNSHIP:	'[' 't' 'w' 'p' ']' ;
UMIL:		'[' 'm' 'i' 'l' '_' 'u' 's' ']' ;
BINCH:		'[' 'i' 'n' '_' 'b' 'r' ']' ;
BFOOT:		'[' 'f' 't' '_' 'b' 'r' ']' ;
BROD:		'[' 'r' 'd' '_' 'b' 'r' ']' ;
BCHAIN:		'[' 'c' 'h' '_' 'b' 'r' ']' ;
BLINK:		'[' 'l' 'k' '_' 'b' 'r' ']' ;
BFATHOM:	'[' 'f' 't' 'h' '_' 'b' 'r' ']' ;
PACE:		'[' 'p' 'c' '_' 'b' 'r' ']' ;
BYARD:		'[' 'y' 'd' '_' 'b' 'r' ']' ;
BMILE:		'[' 'm' 'i' '_' 'b' 'r' ']' ;
BNMILE:		'[' 'n' 'm' 'i' '_' 'b' 'r' ']' ;
BKNOT:		'[' 'k' 'n' '_' 'b' 'r' ']' ;
BACRE:		'[' 'a' 'c' 'r' '_' 'b' 'r' ']' ;
UGALLON:	'[' 'g' 'a' 'l' '_' 'u' 's' ']' ;
UBARREL:	'[' 'b' 'b' 'l' '_' 'u' 's' ']' ;
UQUART:		'[' 'q' 't' '_' 'u' 's' ']' ;
UPINT:		'[' 'p' 't' '_' 'u' 's' ']' ;
UGILL:		'[' 'g' 'i' 'l' '_' 'u' 's' ']' ;
UFLOUNCE:	'[' 'f' 'o' 'z' '_' 'u' 's' ']' ;
UFLDRAM:	'[' 'f' 'd' 'r' '_' 'u' 's' ']' ;
UMINIM:		'[' 'm' 'i' 'n' '_' 'u' 's' ']' ;
UCORD:		'[' 'c' 'r' 'd' '_' 'u' 's' ']' ;
UBUSHEL:	'[' 'b' 'u' '_' 'u' 's' ']' ;
WINGALLON:	'[' 'g' 'a' 'l' '_' 'w' 'i' ']' ;
UPECK:		'[' 'p' 'k' '_' 'u' 's' ']' ;
UDRYQUART:	'[' 'd' 'q' 't' '_' 'u' 's' ']' ;
UDRYPINT:	'[' 'd' 'p' 't' '_' 'u' 's' ']' ;
UTABLESPOON:	'[' 't' 'b' 's' '_' 'u' 's' ']' ;
UTEASPOON:	'[' 't' 's' 'p' '_' 'u' 's' ']' ;
UCUP:		'[' 'c' 'u' 'p' '_' 'u' 's' ']' ;
BGALLON:	'[' 'g' 'a' 'l' '_' 'b' 'r' ']' ;
BPECK:		'[' 'p' 'k' '_' 'b' 'r' ']' ;
BBUSHEL:	'[' 'b' 'u' '_' 'b' 'r' ']' ;
BQUART:		'[' 'q' 't' '_' 'b' 'r' ']' ;
BPINT:		'[' 'p' 't' '_' 'b' 'r' ']' ;
BGILL:		'[' 'g' 'i' 'l' '_' 'b' 'r' ']' ;
BFLOUNCE:	'[' 'f' 'o' 'z' '_' 'b' 'r' ']' ;
BFLDRAM:	'[' 'f' 'd' 'r' '_' 'b' 'r' ']' ;
BMINIM:		'[' 'm' 'i' 'n' '_' 'b' 'r' ']' ;
GRAIN:		'[' 'g' 'r' ']' ;
POUND:		'[' 'l' 'b' '_' 'a' 'v' ']' ;
OUNCE:		'[' 'o' 'z' '_' 'a' 'v' ']' ;
DRAM:		'[' 'd' 'r' '_' 'a' 'v' ']' ;
SCWT:		'[' 's' 'c' 'w' 't' '_' 'a' 'v' ']' ;
LCWT:		'[' 'l' 'c' 'w' 't' '_' 'a' 'v' ']' ;
STON:		'[' 's' 't' 'o' 'n' '_' 'a' 'v' ']' ;
LTON:		'[' 'l' 't' 'o' 'n' '_' 'a' 'v' ']' ;
STONE:		'[' 's' 't' 'o' 'n' 'e' '_' 'a' 'v' ']' ;
PENNYWEIGHT:	'[' 'p' 'w' 't' '_' 't' 'r' ']' ;
TROYOUNCE:	'[' 'o' 'z' '_' 't' 'r' ']' ;
TROYPOUND:	'[' 'l' 'b' '_' 't' 'r' ']' ;
SCRUPLE:	'[' 's' 'c' '_' 'a' 'p' ']' ;
APDRAM:		'[' 'd' 'r' '_' 'a' 'p' ']' ;
APOUNCE:	'[' 'o' 'z' '_' 'a' 'p' ']' ;
APPOUND:	'[' 'l' 'b' '_' 'a' 'p' ']' ;
LINE:		'[' 'l' 'n' 'e' ']' ;
POINT:		'[' 'p' 'n' 't' ']' ;
PICA:		'[' 'p' 'c' 'a' ']' ;
PPOINT:		'[' 'p' 'n' 't' '_' 'p' 'r' ']' ;
PPICA:		'[' 'p' 'c' 'a' '_' 'p' 'r' ']' ;
PIED:		'[' 'p' 'i' 'e' 'd' ']' ;
POUCE:		'[' 'p' 'o' 'u' 'c' 'e' ']' ;
LIGNE:		'[' 'l' 'i' 'g' 'n' 'e' ']' ;
DIDOT:		'[' 'd' 'i' 'd' 'o' 't' ']' ;
CICERO:		'[' 'c' 'i' 'c' 'e' 'r' 'o' ']' ;
FAHRENHEIT:	'[' 'd' 'e' 'g' 'F' ']' ;
CALORIE15:	'c' 'a' 'l' '_' '[' '1' '5' ']' ;
CALORIE20:	'c' 'a' 'l' '_' '[' '2' '0' ']' ;
CALORIEM:	'c' 'a' 'l' '_' 'm' ;
CALORIEIT:	'c' 'a' 'l' '_' 'I' 'T' ;
CALORIETH:	'c' 'a' 'l' '_' 't' 'h' ;
CALORIE:	'c' 'a' 'l' ;
CALORIEN:	'[' 'C' 'a' 'l' ']' ;
BTU39:		'[' 'B' 't' 'u' '_' '3' '9' ']' ;
BTU59:		'[' 'B' 't' 'u' '_' '5' '9' ']' ;
BTU60:		'[' 'B' 't' 'u' '_' '6' '0' ']' ;
BTUM:		'[' 'B' 't' 'u' '_' 'm' ']' ;
BTUIT:		'[' 'B' 't' 'u' '_' 'I' 'T' ']' ;
BTUTH:		'[' 'B' 't' 'u' '_' 't' 'h' ']' ;
BTU:		'[' 'B' 't' 'u' ']' ;
METERH2O:	'm' '[' 'H' '2' 'O' ']' ;
METERMERC:	'm' '[' 'H' 'g' ']' ;
INCHH2O:	'[' 'i' 'n' '_' 'i' '\'' 'H' '2' 'O' ']' ;
INCHMERC:	'[' 'i' 'n' '_' 'i' '\'' 'H' 'g' ']' ;
PERVASCRES:	'[' 'P' 'R' 'U' ']' ;
DIOPTER:	'[' 'd' 'i' 'o' 'p' ']' ;
MESH:		'[' 'm' 'e' 's' 'h' '_' 'i' ']' ;
CHARRIERE:	'[' 'C' 'h' ']' ;
DROP:		'[' 'd' 'r' 'p' ']' ;
HU:		'[' 'h' 'n' 's' 'f' '\'' 'U' ']' ;
EQUIVALENTS:	"eq" ;
OSMOLE:		"osm" ;
ACIDITY:	'[' 'p' 'H' ']' ;
GRAMPERCENT:	"g%" ;
SVEDBERG:	'[' 'S' ']' ;
HIGHPOWERFIELD:	'[' 'H' 'P' 'F' ']' ;
LOWPOWERFIELD:	'[' 'L' 'P' 'F' ']' ;
KATAL:		"kat" ;
UNIT:		"U" ;
INTUNIT:	'[' 'i' 'U' ']' ;
ARBUNIT:	'[' 'a' 'r' 'b' '\'' 'U' ']' ;
USPUNIT:	'[' 'U' 'S' 'P' '\'' 'U' ']' ;
GPLUNIT:	'[' 'G' 'P' 'L' '\'' 'U' ']' ;
MPLUNIT:	'[' 'M' 'P' 'L' '\'' 'U' ']' ;
APLUNIT:	'[' 'A' 'P' 'L' '\'' 'U' ']' ;
BETHESDAUNIT:	'[' 'b' 'e' 't' 'h' '\'' 'U' ']' ;
TODDUNIT:	'[' 't' 'o' 'd' 'd' '\'' 'U' ']' ;
DYEUNIT:	'[' 'd' 'y' 'e' '\'' 'U' ']' ;
SOMOGYIUNIT:	'[' 's' 'm' 'g' 'y' '\'' 'U' ']' ;
BODANSKYUNIT:	'[' 'b' 'd' 's' 'k' '\'' 'U' ']' ;
KIMGARMUNIT:	'[' 'k' 'a' '\'' 'U' ']' ;
KUNKELUNIT:	'[' 'k' 'n' 'k' '\'' 'U' ']' ;
MACLAGANUNIT:	'[' 'm' 'c' 'l' 'g' '\'' 'U' ']' ;
TUBERUNIT:	'[' 't' 'b' '\'' 'U' ']' ;
NEPER:		"Np" ;
BEL:		"B" ;
BELSOUNDPRES:	'B' '[' 'S' 'P' 'L' ']' ;
BELVOLT:	'B' '[' 'V' ']' ;
BELMILLIVOLT:	'B' '[' 'm' 'V' ']' ;
BELMICROVOLT:	'B' '[' 'u' 'V' ']' ;
BELWATT:	'B' '[' 'W' ']' ;
BELKILOWATT:	'B' '[' 'k' 'W' ']' ;
STERE:		"st" ;
ANGSTROM:	"Ao" ;
BARN:		"b" ;
ATMOSPHERE:	"att" ;
PSI:		'[' 'p' 's' 'i' ']' ;
CIRCLE:		"circ" ;
SPHERE:		"sph" ;
CARATM:		'[' 'c' 'a' 'r' '_' 'm' ']' ;
CARATG:		'[' 'c' 'a' 'r' '_' 'A' 'u' ']' ;
BIT_S:		"bit_s" ;
BIT:		"bit" ;
BYTE:		"By" ;
BAUD:		"Bd" ;
KIBI:		"Ki" ;
MEBI:		"Mi" ;
GIBI:		"Gi" ;
TEBI:		"Ti" ;
HORSEPOWER:	'[' 'H' 'P' ']' ;
METABOLICEQUIV:	'[' 'M' 'E' 'T' ']' ;
HOMEOPOTDEC:	'[' 'h' 'p' '_' 'X' ']' ;
HOMEOPOTCEN:	'[' 'h' 'p' '_' 'C' ']' ;
MHO:		"mho" ;

KELVINDELTA:	'[' 'd' 'e' 'g' 'K' '_' 'd' 'e' 'l' 't' 'a' ']' ;
CELCIUSDELTA:	'[' 'd' 'e' 'g' 'C' '_' 'd' 'e' 'l' 't' 'a' ']' ;
FAHRENDELTA:	'[' 'd' 'e' 'g' 'F' '_' 'd' 'e' 'l' 't' 'a' ']' ;

EOL: 		( '\r' | '\n' ) ;    
